/*------------------------------------------------------------------------------
	Newton interface to the native GIF encoding/decoding functions.
------------------------------------------------------------------------------*/

#include "Objects.h"
#include "NewtonScript.h"
#include "GIF.h"


/*------------------------------------------------------------------------------
	C o n s t a n t s
------------------------------------------------------------------------------*/

#define kGIFPreambleLength	82
#define kVBOThreshold		(2*KByte)


/*------------------------------------------------------------------------------
	Return a pointer to the bits in a PixelMap.
	Part of the QuickDraw suite in the Newton ROM.

	Args:		pixmap
					the PixelMap

	Return:	a pointer to the bitmap
------------------------------------------------------------------------------*/

extern "C" Ptr
GetPixelMapBits(PixelMap * pixmap)
{
	Ptr pixBits = pixmap->baseAddr;
	ULong storage = pixmap->pixMapFlags & kPixMapStorage;

	if (storage == kPixMapOffset)
		return pixBits + (unsigned) pixmap;
	else if (storage == kPixMapHandle)
		return *(Handle) pixBits;
	else if (storage == kPixMapPtr)
		return pixBits;
	return NULL;
}


/*------------------------------------------------------------------------------
	Return an options frame sutable for passing to MakeBitmap.
	NOTE	Should check the actual screen depth.

	Args:		inStore
					the store on which to create the bitmap object

	Return:	the options frame ref
------------------------------------------------------------------------------*/

Ref
MakeOptionsFrame(RefArg inStore)
{
	RefVar options(AllocateFrame());

	SetFrameSlot(options, SYM(depth), MAKEINT(4));
	if (NOTNIL(inStore))
		SetFrameSlot(options, SYM(store), inStore);

	return options;
}


/*------------------------------------------------------------------------------
	Decode GIF image.

	Args:		inData
					the GIF data
				inStore
					store on which to create VBO

	Return:	a bitmap binary object
------------------------------------------------------------------------------*/

extern "C"
Ref Decode(RefArg rcvr, RefArg inData, RefArg inStore)
{
	if (!IsBinary(inData))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, inData);

	RefVar bitmap;
	RefVar options(MakeOptionsFrame(inStore));
	CGIFReader * reader = NULL;

	WITH_LOCKED_BINARY(inData, dataPtr)
		UShort width, height;
		reader = new CGIFReader((UChar *) dataPtr, Length(inData));
		if (reader->StartGIF(&width, &height))
		{
			bitmap = NSCallGlobalFn(SYM(MakeBitmap), MAKEINT(width), MAKEINT(height), options);
			if (NOTNIL(bitmap))
			{
			// bitmap.data: <pixels:#x>
			// bitmap.colorData: { cbits: <cbits:#x>, bitDepth: 4, colorTable: <grayPattern:#128> }
				RefVar pix(GetFrameSlot(bitmap, SYM(data)));
				WITH_LOCKED_BINARY(pix, pixPtr)
					reader->ReadGIF((PixelMap *) pixPtr);
				END_WITH_LOCKED_BINARY(pix)
			}
		}
/*	END_WITH_LOCKED_BINARY(inData)
	 but we also ALWAYS want to delete the reader so we expand the macro and include the delete */
#if 0
	{	// balance bracket for source code parser
#endif
	}
	on_unwind {
		UnlockRefArg(inData);
		if (reader)
			delete reader;
	}
	end_unwind;

	return bitmap;
}


/*------------------------------------------------------------------------------
	Encode GIF image.

	Args:		inBitmap
					a bitmap frame
				inStore
					store on which to create VBO

	Return:	a GIF binary object
------------------------------------------------------------------------------*/

extern "C"
Ref Encode(RefArg rcvr, RefArg inBitmap, RefArg inStore)
{
	if (!IsFrame(inBitmap))
		ThrowBadTypeWithFrameData(kNSErrNotAFrame, inBitmap);

	RefVar pix(GetFrameSlot(inBitmap, SYM(data)));
	ULong  dataLength = kGIFPreambleLength + Length(pix);
	RefVar gif;
	if (NOTNIL(inStore) && dataLength > kVBOThreshold)
		gif = NSSend(inStore, SYM(NewVBO), SYM(gif), MAKEINT(dataLength));
	else
		gif = AllocateBinary(SYM(gif), dataLength);

	WITH_LOCKED_BINARY(pix, pixPtr)
		CGIFWriter * writer = NULL;
		WITH_LOCKED_BINARY(gif, dataPtr)
			writer = new CGIFWriter();
			dataLength = writer->WriteGIF((PixelMap *) pixPtr, (UChar *) dataPtr, dataLength);
/*		END_WITH_LOCKED_BINARY(gif)
		 but we also ALWAYS want to delete the writer so we expand the macro and include the delete */
#if 0
	{	// balance bracket for source code parser
#endif
		}
		on_unwind {
			UnlockRefArg(gif);
			if (writer)
				delete writer;
		}
		end_unwind;
	END_WITH_LOCKED_BINARY(pix)

	SetLength(gif, dataLength);
	return gif;
}
