/*----------------------------------------------------------------------
	G I F   D e c o d e r

	Adapted from the versit vCard sample code.
----------------------------------------------------------------------*/

#include "Objects.h"
#include "GIF.h"


/*----------------------------------------------------------------------
	Constructor.
	Initialize all instance variables.
----------------------------------------------------------------------*/

CGIFReader::CGIFReader(UChar * data, size_t length)
{
	dataStart = data;
	dataStop = data + length - 1;
	dataPtr = data;
}


/*----------------------------------------------------------------------
	Destructor.
----------------------------------------------------------------------*/

CGIFReader::~CGIFReader()
{
}


/*----------------------------------------------------------------------
	Start reading data from GIF stream to determine its size.
----------------------------------------------------------------------*/

BOOL
CGIFReader::StartGIF(UShort * width, UShort * height)
{
	dataPtr = dataStart;
	if (ReadHeader())
	{
		ReadLogicalScreenDescriptor();
    	*width = screen.width;
    	*height = screen.height;
    	return YES;
	}
	return NO;
}


/*----------------------------------------------------------------------
	Read data from GIF stream and fill bitmap binary data.
----------------------------------------------------------------------*/

void
CGIFReader::ReadGIF(PixelMap * pixels)
{
	size_t	size;
	BOOL		err;

	pixmap = pixels;
	pixBits = GetPixelMapBits(pixmap);
	pixRowBytes = pixmap->rowBytes & 0x3FFF;

	bitsPerPixel = PixelDepth(pixmap);
	render = (bitsPerPixel < 4) ? &Render1bit : &Render4bit;
	currentColorTable = NULL;
	while (dataPtr < dataStop)
	{
		switch (GetByte())
		{
		case kImageLabel:
			ReadImageDescriptor();
			if (image.hasLocalColorTable)
				currentColorTable = image.localColorTable;
			else
				currentColorTable = screen.globalColorTable;
			lineCount = 0;
			pass = 0;

			// Initialize for decoding a new image...
			size = GetByte();
			if (2 <= size && size <= 9)
			{
				InitDecoder(size);
				err = !Decode(image.width);
			}
			else
				err = YES;
			if (image.hasLocalColorTable)
				delete [] image.localColorTable;
			if (err)
				goto bail;
			break;

		case kExtensionIntro:
			switch (GetByte())
			{
			case kGraphicControlLabel:
				ReadGraphicControlExtension();
				break;

			case kCommentLabel:
				TrashCommentExtension();
				break;

			case kPlainTextLabel:
				TrashPlainTextExtension();
				break;

			case kApplicationLabel:
				TrashApplicationExtension();
				break;
			}
			break;

		case kTrailer:
		//	printf ("clean input stream termination\n");
			break;

		default:
		//	printf ("bad input stream end %d\n", (int)(*dataPtr - 1));
			break;
		}
	}
bail:
	if (screen.hasGlobalColorTable)
		delete [] screen.globalColorTable;
}

#pragma mark -

/*----------------------------------------------------------------------
	B l o c k   R e a d i n g   F u n c t i o n s
----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
	Read GIF file header.

	The Header identifies the GIF Data Stream in context. The sig[nature] field
	marks the beginning of the Data Stream, and the version field identifies the
	set of capabilities required of a decoder to fully process the Data Stream.

	Signature :										"GIF"
	Version Numbers as of 10 July 1990 :	"87a" - May 1987
														"89a" - July 1989 

	Args:

	Return:
		YES	if header is valid
		NO		if it isnt
----------------------------------------------------------------------*/

BOOL
CGIFReader::ReadHeader()
{
	char	sig[6];

	for (int i = 0; i < 6; i++)
		sig[i] = GetByte();

	return ( sig[0] == 'G'
			&& sig[1] == 'I'
			&& sig[2] == 'F'
			&& sig[3] == '8' );
}


/*----------------------------------------------------------------------
	Read GIF colour table.

	Args:
		count	number of entries to read
		ct		pointer to RGB colour table
	Return:
		void
----------------------------------------------------------------------*/

void
CGIFReader::ReadColorTable(UShort count, GIFRGBPtr ct)
{
	for (int i = 0; i < count; i++)
	{
		ct->r = GetByte();
		ct->g = GetByte();
		ct->b = GetByte();
		ct++;
	}
}


/*----------------------------------------------------------------------
	Read GIF logical screen descriptor.

	Args:
		screen	pointer to logical screen descriptor
	Return:
		void
----------------------------------------------------------------------*/

void
CGIFReader::ReadLogicalScreenDescriptor()
{
	// Read logical screen dimensions
	screen.width = GetWord();
	screen.height = GetWord();

	// Read flags
	screen.flags = GetByte();
	// Unpack them
	screen.hasGlobalColorTable = screen.flags >> 7;
	screen.colorResolution = (screen.flags >> 4) & 0x07;
	screen.isSorted = (screen.flags >> 3) & 0x01;
	screen.globalColorTableSize = 1 << ((screen.flags & 0x07) + 1);

	// Read background colour
	screen.backgroundColor = GetByte();
	// Read aspect ratio
	screen.aspectRatio = GetByte();

	// Create global colour table
	if (screen.hasGlobalColorTable)
	{
		screen.globalColorTable = new GIFRGB[screen.globalColorTableSize];
		ReadColorTable(screen.globalColorTableSize, screen.globalColorTable);
	}
	else
		screen.globalColorTable = NULL;
}


/*----------------------------------------------------------------------
	Read GIF image descriptor.

	Args:
		image	pointer to image descriptor
	Return:
		void
----------------------------------------------------------------------*/

void
CGIFReader::ReadImageDescriptor()
{
	// Read image dimensions
	image.left = GetWord();
	image.top = GetWord();
	image.width = GetWord();
	image.height = GetWord();

	// Read flags
	image.flags = GetByte();
	// Unpack them
	image.hasLocalColorTable = image.flags >> 7;
	image.isInterlaced = (image.flags >> 6) & 0x01;
	image.isSorted = (image.flags >> 5) & 0x01;
	image.localColorTableSize = 1 << ((image.flags & 0x07) + 1);

	// Create local colour table
	if (image.hasLocalColorTable)
	{
		image.localColorTable = new GIFRGB[image.localColorTableSize];
		ReadColorTable(image.localColorTableSize, image.localColorTable);
	}	
	else
		image.localColorTable = NULL;
}


/*----------------------------------------------------------------------
	Read GIF graphic control extension.

	Args:
		ext	pointer to control block
	Return:
		YES	if block is valid
		NO		if it isnt
----------------------------------------------------------------------*/

BOOL
CGIFReader::ReadGraphicControlExtension()
{
	UChar b;

	// Validate block size
	if ((b = GetByte()) != kGraphicControlBlockSize)
		return NO;
	
	// Read flags
	gext.flags = GetByte();
	// Unpack them
	gext.hasTransparency = (gext.flags & 0x01);

	// Read delay time
	gext.delayTime = GetWord();

	// Read transparent colour index
	gext.transparentColor = GetByte();

	// Validate block terminator
	return (b = GetByte()) == kBlockTerminator;
}


/*----------------------------------------------------------------------
	B l o c k   S k i p p i n g   ( T r a s h i n g )   F u n c t i o n s
----------------------------------------------------------------------*/

void
CGIFReader::TrashDataSubBlocks()
{
	UByte b;

	while ((b = GetByte()) != kBlockTerminator)
	{
		Advance(b);
	}
}


void
CGIFReader::TrashCommentExtension()
{
	TrashDataSubBlocks();
}


void
CGIFReader::TrashPlainTextExtension()
{
	Advance(kPlainTextBlockSize);
	TrashDataSubBlocks();
}


void
CGIFReader::TrashApplicationExtension()
{
	Advance(kApplicationBlockSize);
	TrashDataSubBlocks();
}


/*----------------------------------------------------------------------
	I n p u t   s t r e a m
------------------------------------------------------------------------
	Read bytes, words and nothingness.
----------------------------------------------------------------------*/

UChar
CGIFReader::GetByte(void)
{
	if (dataPtr > dataStop)
		ThrowBadTypeWithFrameData(kNSErrOutOfRange, MAKEINT(dataPtr - dataStart));
	return *dataPtr++;
}


UShort
CGIFReader::GetWord(void)
{
	UShort w = GetByte();
	return w + (GetByte() << 8);
}


void
CGIFReader::Advance(int count)
{
	dataPtr += count;
}


/*----------------------------------------------------------------------
	Get the next byte from the GIF stream.
	Must handle the block structured nature of the stream.
----------------------------------------------------------------------*/

UChar
CGIFReader::GetBlockByte(void)
{
	if (availableBytes == 0)
		// Out of bytes in current block, so read next block
		availableBytes = GetByte();	// number of bytes in following block

	availableBytes--;
	return GetByte();
}


/*----------------------------------------------------------------------
	Get the next code from the GIF stream.
----------------------------------------------------------------------*/

ULong
CGIFReader::GetCode(void)
{
	const ULong codeMask[13] =
	{
		0,
		0x0001, 0x0003, 0x0007, 0x000F,
		0x001F, 0x003F, 0x007F, 0x00FF,
		0x01FF, 0x03FF, 0x07FF, 0x0FFF
	};

	ULong code;

	if (availableBits == 0)
	{ 
		currentByte = GetBlockByte();
		availableBits = 8;
	}

	code = currentByte >> (8 - availableBits);

	while (currentCodeSize > availableBits)
	{
		currentByte = GetBlockByte();
		code |= (currentByte << availableBits);
		availableBits += 8;
	}

	availableBits -= currentCodeSize;
	return code & codeMask[currentCodeSize];
}

#pragma mark -

/*----------------------------------------------------------------------
	I m a g e   D e c o d i n g   F u n c t i o n s
----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
	Initialize the decoder for reading a new image.
----------------------------------------------------------------------*/

void
CGIFReader::InitDecoder(size_t size)
{
	currentByte = 0;
	availableBytes = 0;
	availableBits = 0;

	initialCodeSize = size;
	clearCode = 1 << size;
	endCode = clearCode + 1;
	newCodes = clearCode + 2;

	ResetDecoder();
}


void
CGIFReader::ResetDecoder(void)
{
	currentCodeSize = initialCodeSize + 1;
	slot = newCodes;
	topSlot = 1 << currentCodeSize;
}


/* bool decoder(linewidth)
 *    short linewidth;               * Pixels per line of image *
 *
 * - This function decodes an LZW image, according to the method used
 * in the GIF spec.  Every *linewidth* "characters" (ie. pixels) decoded
 * will generate a call to OutLine(), which is a user specific function
 * to display a line of pixels.  The function gets it's codes from
 * get_next_code() which is responsible for reading blocks of data and
 * seperating them into the proper size codes.  Finally, get_byte() is
 * the global routine to read the next byte from the GIF stream.
 *
 * It is generally a good idea to have linewidth correspond to the actual
 * width of a line (as specified in the Image header) to make your own
 * code a bit simpler, but it isn't absolutely necessary.
 *
 * Returns: true if successful, else false.
 */

BOOL
CGIFReader::Decode(UShort linewidth)
{
   UChar *	sp;
   UChar *	bufPtr;
   UChar *	buf;
   ULong		fc, oc;
   ULong		c, code, bufCnt;
   UChar		b;
   BOOL		success = NO;

	// Initialize in case they forgot to put in a clear code.
	// (This shouldn't happen, but we'll try and decode it anyway...)
	oc = fc = 0;

	// Allocate space for the decode buffer
	buf = (UChar *) malloc(linewidth + 1);
	if (buf != NULL)
	{
		// Set up the stack pointer and decode buffer pointer
		bufPtr = buf;
		bufCnt = linewidth;
		sp = stack;
	
		/* This is the main loop.  For each code we get we pass through the
		* linked list of prefix codes, pushing the corresponding "character" for
		* each code onto the stack.  When the list reaches a single "character"
		* we push that on the stack too, and then start unstacking each
		* character for output in the correct order.  Special handling is
		* included for the clear code, and the whole thing ends when we get
		* an ending code.
		*/
		while ((c = GetCode()) != endCode)
		{
			// If the code is a clear code, reinitialize all necessary items.
			if (c == clearCode)
			{
				ResetDecoder();
	
				/* Continue reading codes until we get a non-clear code
				* (Another unlikely, but possible case...)
				*/
				while ((c = GetCode()) == clearCode)
					;
	
				/* If we get an ending code immediately after a clear code
				* (Yet another unlikely case), then break out of the loop.
				*/
				if (c == endCode)
					break;
	
				/* Finally, if the code is beyond the range of already set codes,
				* (This one had better NOT happen...  I have no idea what will
				* result from this, but I doubt it will look good...) then set it
				* to color zero.
				*/
				if (c >= slot)
					c = 0;
	
				oc = fc = c;
	
				/* And let us not forget to put the char into the buffer... And
				* if, on the off chance, we were exactly one pixel from the end
				* of the line, we have to send the buffer to the OutputLine()
				* routine...
				*/
				*bufPtr++ = (UChar) c;
				if (--bufCnt == 0)
				{
					OutputLine(buf, linewidth);
					bufPtr = buf;
					bufCnt = linewidth;
				}
			}
			else
			{
				/* In this case, it's not a clear code or an ending code, so
				* it must be a code code...  So we can now decode the code into
				* a stack of character codes. (Clear as mud, right?)
				*/
				code = c;

				/* Here we go again with one of those off chances...  If, on the
				* off chance, the code we got is beyond the range of those already
				* set up (Another thing which had better NOT happen...) we trick
				* the decoder into thinking it actually got the last code read.
				* (Hmmn... I'm not sure why this works...  But it does...)
				*/
				if (code >= slot)
				{
					code = oc;
					*sp++ = (UChar) fc;
				}

				/* Here we scan back along the linked list of prefixes, pushing
				* helpless characters (ie. suffixes) onto the stack as we do so.
				*/
				while (code >= newCodes)
				{
					*sp++ = suffix[code];
					code = prefix[code];
				}

				/* Push the last character on the stack, and set up the new
				* prefix and suffix, and if the required slot number is greater
				* than that allowed by the current bit size, increase the bit
				* size.  (NOTE - If we are all full, we *don't* save the new
				* suffix and prefix...  I'm not certain if this is correct...
				* It might be more proper to overwrite the last code...
				*/
				*sp++ = (UChar) code;
				if (slot < topSlot)
				{
					fc = code;
					suffix[slot] = (UChar) fc;
					prefix[slot++] = (UShort) oc;
					oc = c;
				}

				if (slot >= topSlot)
				{
					if (currentCodeSize < GIFBITS)
					{
						topSlot <<= 1;
						++currentCodeSize;
					}
				}

				/* Now that we've pushed the decoded string (in reverse order)
				* onto the stack, let's pop it off and put it into our decode
				* buffer...
				* And when the decode buffer is full, write another line...
				*/
				while (sp > stack)
				{
					*bufPtr++ = *(--sp);
					if (--bufCnt == 0)
					{
						OutputLine(buf, linewidth);
						bufPtr = buf;
						bufCnt = linewidth;
					}
				}
			}
		}
	
		if (bufCnt != linewidth)
		{
			OutputLine(buf, (linewidth - bufCnt));
		}
	
		// Skip block terminator
		success = ((b = GetByte()) == kBlockTerminator);
	}
	free(buf);
	return success;
}

#pragma mark -

/*----------------------------------------------------------------------
	O u t p u t   F u n c t i o n s
----------------------------------------------------------------------*/

/* Reduce RGB values to grayscale, with or without alpha
 * using the equation given in Poynton's ColorFAQ at
 * <http://www.inforamp.net/~poynton/>
 * Copyright (c) 1998-01-04 Charles Poynton poynton@inforamp.net
 *
 *     Y = 0.212671 * R + 0.715160 * G + 0.072169 * B
 *
 *  We approximate this with
 *
 *     Y = 0.21268 * R    + 0.7151 * G    + 0.07217 * B
 *
 *  which can be expressed with integers as
 *
 *     Y = (6969 * R + 23434 * G + 2365 * B)/32768
 *
 *  The calculation is to be done in a linear colorspace.
 */

inline UChar
MapRGB(GIFRGBPtr pRGB)
{
	return ~((6969 * (unsigned) pRGB->r
			 + 23434 * (unsigned) pRGB->g
			 +  2365 * (unsigned) pRGB->b) >> 15);
}


void
CGIFReader::OutputLine(UChar * line, size_t length)
{
	UChar *	pixPtr = (UChar *) pixBits + pixRowBytes * lineCount;

	(*render)(line, length, currentColorTable, pixPtr);

	if (image.isInterlaced)
		switch (pass)
		{
		case 0:
			lineCount += 8;
			if (lineCount >= (image.top + image.height))
			{
				pass++;
				lineCount = image.top + 4;
			}
    		break;
		case 1:
			lineCount += 8;
			if (lineCount >= (image.top + image.height))
			{
				pass++;
				lineCount = image.top + 2;
			}
    		break;
		case 2:
			lineCount += 4;
			if (lineCount >= (image.top + image.height))
			{
				pass++;
				lineCount = image.top + 1;
			}
    		break;
    	case 3:
			lineCount += 2;
		}
	else	
		lineCount++;
}


void
CGIFReader::Render4bit(UChar * line, size_t length, GIFRGBPtr colorTable, UChar * pixPtr)
{
	UChar		pixHi, pixLo;
	size_t	i;

	for (i = length/2; i > 0; i--)
	{
		pixHi = MapRGB(&(colorTable[*line++])) & 0xF0;
		pixLo = MapRGB(&(colorTable[*line++])) >> 4;
		*pixPtr++ = pixHi | pixLo;
	}
	if (length & 1)	// if width is odd, write out odd pixel
		*pixPtr = MapRGB(&(colorTable[*line++])) & 0xF0;
}


void
CGIFReader::Render1bit(UChar * line, size_t length, GIFRGBPtr colorTable, UChar * pixPtr)
{
	UChar		pix3, pix2, pix1, pix0;
	UChar		pixHi, pixLo;
	size_t	i, j;

	for (i = length; i >= 8; i -= 8)
	{
		pix3 = MapRGB(&(colorTable[*line++])) & 0x80;
		pix2 = (MapRGB(&(colorTable[*line++])) >> 1) & 0x40;
		pix1 = (MapRGB(&(colorTable[*line++])) >> 2) & 0x20;
		pix0 = (MapRGB(&(colorTable[*line++])) >> 3) & 0x10;
		pixHi = pix3 | pix2 | pix1 | pix0;
		pix3 = (MapRGB(&(colorTable[*line++])) >> 4) & 0x08;
		pix2 = (MapRGB(&(colorTable[*line++])) >> 5) & 0x04;
		pix1 = (MapRGB(&(colorTable[*line++])) >> 6) & 0x02;
		pix0 = (MapRGB(&(colorTable[*line++])) >> 7) & 0x01;
		pixLo = pix3 | pix2 | pix1 | pix0;
		*pixPtr++ = pixHi | pixLo;
	}
	if (i > 0)
	{
		if (i >= 4)
		{
			pix3 = MapRGB(&(colorTable[*line++])) & 0x80;
			pix2 = (MapRGB(&(colorTable[*line++])) >> 1) & 0x40;
			pix1 = (MapRGB(&(colorTable[*line++])) >> 2) & 0x20;
			pix0 = (MapRGB(&(colorTable[*line++])) >> 3) & 0x10;
			pixHi = pix3 | pix2 | pix1 | pix0;
			j = 4;
		}
		else
		{
			pixHi = 0;
			j = 0;
		}
		pixLo = 0;
		for ( ; j < i; j++)
			pixLo |= ((MapRGB(&(colorTable[*line++])) >> j) & (0x80 >> j));
		*pixPtr = pixHi | pixLo;
	}
}
