/*----------------------------------------------------------------------
	G I F   E n c o d e r

	Adapted from GIFENCODE by David Rowley.
	Lempel-Ziv compression based on compress.c - file compression
	 la IEEE Computer, June 1984.
----------------------------------------------------------------------*/

#include "Objects.h"
#include "GIF.h"

#define backgroundColor   0
#define transparentColor -1


/*----------------------------------------------------------------------
	Constructor.
	Initialize all instance variables.
----------------------------------------------------------------------*/

CGIFWriter::CGIFWriter()
{
	isInterlaced = NO;
}


/*----------------------------------------------------------------------
	Destructor.
----------------------------------------------------------------------*/

CGIFWriter::~CGIFWriter()
{
}


/*----------------------------------------------------------------------
	Read data from pixel map and write GIF stream.
----------------------------------------------------------------------*/

size_t
CGIFWriter::WriteGIF(const PixelMap * pixels, UChar * data, size_t length)
{
	pixmap = pixels;
	pixBits = GetPixelMapBits((PixelMap *) pixmap);
	pixRowBytes = pixmap->rowBytes & 0x7FFF;
	pixWidth = pixmap->bounds.right - pixmap->bounds.left;
	pixHeight = pixmap->bounds.bottom - pixmap->bounds.top;

	bitsPerPixel = PixelDepth(pixmap);
	resolution = bitsPerPixel;

	dataStart = data;
	dataStop = data + length - 1;
	dataPtr = data;

	/*
	 * Write the GIF image file.
	 */
	WriteHeader();
	WriteLogicalScreenDescriptor();
	WriteGlobalColorTable();
	WriteGraphicControlExtension();
	WriteImageDescriptor();
	WriteImage();
	WriteTrailer();

	return dataPtr - dataStart;	// return actual length of output stream
}


/*----------------------------------------------------------------------
	I n p u t   s t r e a m
------------------------------------------------------------------------
	Return the next pixel from the image.
----------------------------------------------------------------------*/

int
CGIFWriter::GetNextPixel(void)
{
	UChar *	pixPtr;
	int		pix;

	if (pixLoc.v >= pixHeight)
		return EOF;

	if (bitsPerPixel == 4)
	{
		pixPtr = (UChar *) pixBits + (pixRowBytes * pixLoc.v) + (pixLoc.h >> 1);
		if (pixLoc.h & 0x01)
			pix = *pixPtr & 0x0F;
		else
			pix = *pixPtr >> 4;
	}
	else	// bitsPerPixel must be 1
	{
		pixPtr = (UChar *) pixBits + (pixRowBytes * pixLoc.v) + (pixLoc.h >> 3);
		pix = (*pixPtr >> (pixLoc.h & 0x07)) & 0x01;
	}

	/*
 	* Bump the current X position
 	*/
	pixLoc.h++;

	/*
 	* If we are at the end of a scan line, set pixLoc.h back to the beginning
 	* If we are interlaced, bump the pixLoc.v to the appropriate spot,
 	* otherwise, just increment it.
 	*/
	if (pixLoc.h >= pixWidth)
	{
		pixLoc.h = 0;

		if (isInterlaced)
     		switch (pass)
			{
			case 0:
				pixLoc.v += 8;
				if (pixLoc.v >= pixHeight)
				{
					pass++;
					pixLoc.v = 4;
				}
				break;

			case 1:
				pixLoc.v += 8;
				if (pixLoc.v >= pixHeight)
				{
					pass++;
					pixLoc.v = 2;
				}
				break;

			case 2:
				pixLoc.v += 4;
				if (pixLoc.v >= pixHeight)
				{
					pass++;
					pixLoc.v = 1;
				}
				break;

			case 3:
				pixLoc.v += 2;
				break;
			}
		else
			pixLoc.v++;
	}

	return pix;
}

#pragma mark -

/*----------------------------------------------------------------------
	C o m p r e s s i o n
------------------------------------------------------------------------
	Encode image as GIF file.
----------------------------------------------------------------------*/

void
CGIFWriter::InitEncoder(size_t size)
{
	currentByte = 0;
	usedBits = 0;
	blockSize = 0;

	initialCodeSize = size;
	clearCode = 1 << size;
	endCode = clearCode + 1;
	newCodes = clearCode + 2;

	ResetEncoder();
}


void
CGIFWriter::ResetEncoder(void)
{
	currentCodeSize = initialCodeSize + 1;
	slot = newCodes;
	topSlot = 1 << currentCodeSize;

	memset(hashTable, -1, sizeof(hashTable));
	PutCode(clearCode);
}


/*----------------------------------------------------------------------
	Write compressed GIF image data.

	Algorithm:  use open addressing double hashing (no chaining) on the
	prefix code / next character combination.  We do a variant of Knuth's
	algorithm D (vol. 3, sec. 6.4) along with G. Knott's relatively-prime
	secondary probe.  Here, the modular division first probe gives way
	to a faster exclusive-or manipulation.  Also do block compression with
	an adaptive reset, whereby the code table is cleared when the compression
	ratio decreases, but after the table fills.  The variable-length output
	codes are re-sized at this point, and a special CLEAR code is generated
	for the decompressor.
----------------------------------------------------------------------*/

void
CGIFWriter::Encode(void)
{
	register long fcode;
	register int code1, code2;
	register int i;
	register int delta;

	pass = pixLoc.h = pixLoc.v = 0;		// init pixmap position
	code1 = GetNextPixel();
	while ((code2 = GetNextPixel()) != EOF)
	{
		fcode = (code2 << GIFBITS) + code1;
		i = (code2 << HASH_BITS) ^ code1;	// XOR hashing

		if (hashTable[i] == fcode)			// pattern matches
		{
			code1 = codeTable[i];			// so use that code
			continue;
		}

		else if (hashTable[i] >= 0)		// slot contains a pattern
		{
			delta = (i == 0) ? 1 : HASH_SIZE - i;	// secondary hash (after G. Knott)
	probe:
			if ((i -= delta) < 0)
				i += HASH_SIZE;
			if (hashTable[i] == fcode)		// pattern matches
			{
				code1 = codeTable[i];		// so use that code
				continue;
			}
			if (hashTable[i] >= 0)			// should be > ?
				goto probe;
		}
		// else slot is empty

		PutCode(code1);						// emit previous code
		code1 = code2;							// cycle codes

		/*
		* If the next entry is going to be too big for the code size,
		* then increase it if possible.
		*/
		if (slot >= topSlot)
		{
			if (currentCodeSize < GIFBITS)
			{
				topSlot <<= 1;
				++currentCodeSize;
			}
		}
		if (slot < MAX_CODES)			// add current pattern to hash table
		{
			hashTable[i] = fcode;
			codeTable[i] = slot++;
		}
		else
			ResetEncoder();
	}

	PutCode(code1);		// emit final code
}

#pragma mark -

/*----------------------------------------------------------------------
	O u t p u t   s t r e a m
------------------------------------------------------------------------
	Write blocks.
----------------------------------------------------------------------*/

void
CGIFWriter::WriteHeader(void)
{
	const char * sig;
	if (transparentColor >= 0)
		sig = "GIF89a";
	else
		sig = "GIF87a";
	for (int i = 0; i < 6; i++)
		PutByte(sig[i]);
}


void
CGIFWriter::WriteLogicalScreenDescriptor(void)
{
	PutWord(pixWidth);		// Screen Width
	PutWord(pixHeight);		// Screen Height

	/*
	 * Packed fields.
	 * Indicate that there is a global colour map.
	 */
	PutByte(0x80 | ((resolution - 1) << 4) | (bitsPerPixel - 1));

	PutByte(backgroundColor);
	PutByte(0);					// Aspect Ratio

}


void
CGIFWriter::WriteGlobalColorTable(void)
{
	UChar tone;
	int colorMapSize = (1 << bitsPerPixel) - 1;
	for (int i = colorMapSize; i >= 0; i--)
	{
		tone = 0xFF * i / colorMapSize;
		PutByte(tone);			// Red
		PutByte(tone);			// Green
		PutByte(tone);			// Blue
	}

}


void
CGIFWriter::WriteGraphicControlExtension(void)
{
	if (transparentColor >= 0)
	{
		PutByte(kExtensionIntro);
		PutByte(kGraphicControlLabel);
		PutByte(kGraphicControlBlockSize);
		PutByte(0x01);			// Packed Fields = Transparent Color Flag
		PutWord(0);				// Delay Time
		PutByte(transparentColor);
		PutByte(kBlockTerminator);
   }
}


void
CGIFWriter::WriteImageDescriptor(void)
{
	PutByte(kImageLabel);
	PutWord(0);					// Left Position
	PutWord(0);					// Top Position
	PutWord(pixWidth);		// Image Width
	PutWord(pixHeight);		// Image Height
	PutByte(isInterlaced ? 0x40 : 0x00);	// Packed Fields
}


/*----------------------------------------------------------------------
	Write compressed GIF image data.
----------------------------------------------------------------------*/

void
CGIFWriter::WriteImage(void)
{
	InitEncoder(MAX(2, bitsPerPixel));

	PutByte(initialCodeSize);
	Encode();
	PutCode(endCode);
}


void
CGIFWriter::WriteTrailer(void)
{
	PutByte(kBlockTerminator);
	PutByte(kTrailer);		// the GIF file terminator
}


/*----------------------------------------------------------------------
	Write bytes, words and codes.
----------------------------------------------------------------------*/

void
CGIFWriter::PutByte(UChar byte)
{
	if (dataPtr > dataStop)
		ThrowBadTypeWithFrameData(kNSErrOutOfRange, MAKEINT(dataPtr - dataStart));
	*dataPtr++ = byte;
}


void
CGIFWriter::PutWord(UShort word)
{
	PutByte(word & 0x00FF);
	PutByte(word >> 8);
}


/*----------------------------------------------------------------------
	Put a variable-bit-length code to the GIF stream.
----------------------------------------------------------------------*/

void
CGIFWriter::PutCode(int code)
{
	if (usedBits == 0)
		currentByte = code;
	else
		currentByte |= (code << usedBits);

	usedBits += currentCodeSize;

	while (usedBits >= 8)
	{
		PutBlockByte(currentByte);
		currentByte >>= 8;
		usedBits -= 8;
	}

	if (code == endCode)		// At the end write the rest of the buffer
	{
		while (usedBits > 0)
		{
			PutBlockByte(currentByte);
			currentByte >>= 8;
			usedBits -= 8;
		}
		FlushBlock();
	}
}


/*----------------------------------------------------------------------
	Put the next byte to the GIF stream.
	Must handle the block structured nature of the stream.
----------------------------------------------------------------------*/

void
CGIFWriter::PutBlockByte(UChar byte)
{
	block[blockSize++] = byte;
	if (blockSize == 255)
		FlushBlock();
}


void
CGIFWriter::FlushBlock(void)
{
	if (blockSize > 0)
	{
		UChar *	p = block;
		int		i = blockSize;
		blockSize = 0;
		PutByte(i);
		while (i-- > 0)
			PutByte(*p++);
	}
}
