/* $Header: /d1/sam/tiff/libtiff/RCS/tif_getimage.c,v 1.43 1997/08/29 21:45:52 sam Exp $ */

/*
 * Copyright (c) 1991-1997 Sam Leffler
 * Copyright (c) 1991-1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
 * TIFF Library
 *
 * Read and return a packed RGBA image.
 */
#include "tiffiop.h"
#include <assert.h>
#include <stdio.h>

#define MIN(a,b) ((a)<(b))?(a):(b)

static	int gtTileContig(TIFFRGBAImage *, u_char *, uint32, uint32, uint32);
static	int gtTileSeparate(TIFFRGBAImage *, u_char *, uint32, uint32, uint32);
static	int gtStripContig(TIFFRGBAImage *, u_char *, uint32, uint32, uint32);
static	int gtStripSeparate(TIFFRGBAImage *, u_char *, uint32, uint32, uint32);
static	int pickTileContigCase(TIFFRGBAImage *);
static	int pickTileSeparateCase(TIFFRGBAImage *);

static	const char photoTag[] = "PhotometricInterpretation";


static /*inline*/ u_char
MapRGB(uint32 r, uint32 g, uint32 b)
{
	return (u_char) (uint32) ((6969 * r + 23434 * g +  2365 * b) >> 15);
}

static /*inline*/ u_char
MapRGBA(uint32 r, uint32 g, uint32 b, uint32 a)
{
	return (u_char) (uint32) (((6969 * r + 23434 * g +  2365 * b) >> 15) * a / 255);
}


/*
 * Check the image to see if TIFFReadRGBAImage can deal with it.
 * 1/0 is returned according to whether or not the image can
 * be handled.  If 0 is returned, emsg contains the reason
 * why it is being rejected.
 */
int
TIFFRGBAImageOK(TIFF* tif, char * emsg)
{
	TIFFDirectory* td = &tif->tif_dir;
	uint16 photometric;
	int colorchannels;

	switch (td->td_bitspersample)
	{
	case 1: case 2: case 4: case 8: case 16:
		break;
	default:
		sprintf(emsg, "Sorry, can not handle images with %d-bit samples", td->td_bitspersample);
		return 0;
	}

	colorchannels = td->td_samplesperpixel - td->td_extrasamples;
	if (!TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric))
	{
		switch (colorchannels)
		{
		case 1:
			photometric = PHOTOMETRIC_MINISBLACK;
			break;
		case 3:
			photometric = PHOTOMETRIC_RGB;
			break;
		default:
			sprintf(emsg, "Missing needed %s tag", photoTag);
			return 0;
		}
	}

	switch (photometric)
	{
	case PHOTOMETRIC_MINISWHITE:
	case PHOTOMETRIC_MINISBLACK:
	case PHOTOMETRIC_PALETTE:
		if (td->td_planarconfig == PLANARCONFIG_CONTIG && td->td_samplesperpixel != 1)
		{
			sprintf(emsg, "Sorry, can not handle contiguous data with %s=%d, and %s=%d",
								photoTag, photometric,
								"Samples/pixel", td->td_samplesperpixel);
			return 0;
		}
		break;
	case PHOTOMETRIC_YCBCR:
		if (td->td_planarconfig != PLANARCONFIG_CONTIG)
		{
			sprintf(emsg, "Sorry, can not handle YCbCr images with %s=%d",
								"Planarconfiguration", td->td_planarconfig);
			return 0;
		}
		break;
	case PHOTOMETRIC_RGB: 
		if (colorchannels < 3)
		{
			sprintf(emsg, "Sorry, can not handle RGB image with %s=%d",
								"Color channels", colorchannels);
			return 0;
		}
		break;
#ifdef CMYK_SUPPORT
	case PHOTOMETRIC_SEPARATED:
		if (td->td_inkset != INKSET_CMYK)
		{
			sprintf(emsg, "Sorry, can not handle separated image with %s=%d",
								"InkSet", td->td_inkset);
			return 0;
		}
		if (td->td_samplesperpixel != 4)
		{
			sprintf(emsg, "Sorry, can not handle separated image with %s=%d",
								"Samples/pixel", td->td_samplesperpixel);
			return 0;
		}
		break;
#endif
	case PHOTOMETRIC_LOGL:
		if (td->td_compression != COMPRESSION_SGILOG)
		{
			sprintf(emsg, "Sorry, LogL data must have %s=%d",
								"Compression", COMPRESSION_SGILOG);
			return 0;
		}
		break;
	case PHOTOMETRIC_LOGLUV:
		if (td->td_compression != COMPRESSION_SGILOG
		&& td->td_compression != COMPRESSION_SGILOG24)
		{
			sprintf(emsg, "Sorry, LogLuv data must have %s=%d or %d",
								"Compression", COMPRESSION_SGILOG, COMPRESSION_SGILOG24);
			return 0;
		}
		if (td->td_planarconfig != PLANARCONFIG_CONTIG)
		{
			sprintf(emsg, "Sorry, can not handle LogLuv images with %s=%d",
								"Planarconfiguration", td->td_planarconfig);
			return 0;
		}
		break;
	default:
		sprintf(emsg, "Sorry, can not handle image with %s=%d",
							photoTag, photometric);
		return 0;
	}
	return 1;
}

void
TIFFRGBAImageEnd(TIFFRGBAImage* img)
{
	if (img->Map)
		_TIFFfree(img->Map), img->Map = NULL;
	if (img->BWmap)
		_TIFFfree(img->BWmap), img->BWmap = NULL;
	if (img->PALmap)
		_TIFFfree(img->PALmap), img->PALmap = NULL;
	if (img->ycbcr)
		_TIFFfree(img->ycbcr), img->ycbcr = NULL;
}

static int
IsCCITTCompression(TIFF* tif)
{
	uint16 compress;
	TIFFGetField(tif, TIFFTAG_COMPRESSION, &compress);
	return (compress == COMPRESSION_CCITTFAX3
		  || compress == COMPRESSION_CCITTFAX4
		  || compress == COMPRESSION_CCITTRLE
		  || compress == COMPRESSION_CCITTRLEW);
}


int
TIFFRGBAImageBegin(TIFFRGBAImage* img, TIFF* tif, int stop, char * emsg)
{
	uint16 *	sampleinfo;
	uint16	extrasamples;
	uint16	planarconfig;
	uint16	compress;
	int		colorchannels;

	img->tif = tif;
	img->stoponerr = stop;
	TIFFGetFieldDefaulted(tif, TIFFTAG_BITSPERSAMPLE, &img->bitspersample);
	switch (img->bitspersample)
	{
	case 1: case 2: case 4: case 8: case 16:
		break;
	default:
		sprintf(emsg, "Sorry, can not image with %d-bit samples",
							img->bitspersample);
		return 0;
	}
	img->alpha = 0;
	TIFFGetFieldDefaulted(tif, TIFFTAG_SAMPLESPERPIXEL, &img->samplesperpixel);
	TIFFGetFieldDefaulted(tif, TIFFTAG_EXTRASAMPLES, &extrasamples, &sampleinfo);
	if (extrasamples == 1)
		switch (sampleinfo[0])
		{
		case EXTRASAMPLE_ASSOCALPHA:	/* data is pre-multiplied */
		case EXTRASAMPLE_UNASSALPHA:	/* data is not pre-multiplied */
			img->alpha = sampleinfo[0];
			break;
		}
	colorchannels = img->samplesperpixel - extrasamples;
	TIFFGetFieldDefaulted(tif, TIFFTAG_COMPRESSION, &compress);
	TIFFGetFieldDefaulted(tif, TIFFTAG_PLANARCONFIG, &planarconfig);
	if (!TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &img->photometric))
	{
		switch (colorchannels)
		{
		case 1:
			if (IsCCITTCompression(tif))
				img->photometric = PHOTOMETRIC_MINISWHITE;
			else
				img->photometric = PHOTOMETRIC_MINISBLACK;
			break;
		case 3:
			img->photometric = PHOTOMETRIC_RGB;
			break;
		default:
			sprintf(emsg, "Missing needed %s tag", photoTag);
			return 0;
		}
	}
	switch (img->photometric)
	{
	case PHOTOMETRIC_PALETTE:
		if (!TIFFGetField(tif, TIFFTAG_COLORMAP, &img->redcmap, &img->greencmap, &img->bluecmap))
		{
			TIFFError(TIFFFileName(tif), "Missing required \"Colormap\" tag");
			return 0;
		}
	/* fall thru... */
	case PHOTOMETRIC_MINISWHITE:
	case PHOTOMETRIC_MINISBLACK:
		if (planarconfig == PLANARCONFIG_CONTIG && img->samplesperpixel != 1)
		{
			sprintf(emsg, "Sorry, can not handle contiguous data with %s=%d, and %s=%d",
								photoTag, img->photometric,
								"Samples/pixel", img->samplesperpixel);
			return 0;
		}
		break;
	case PHOTOMETRIC_YCBCR:
		if (planarconfig != PLANARCONFIG_CONTIG)
		{
			sprintf(emsg, "Sorry, can not handle YCbCr images with %s=%d",
								"Planarconfiguration", planarconfig);
			return 0;
		}
		/* It would probably be nice to have a reality check here. */
		if (compress == COMPRESSION_JPEG && planarconfig == PLANARCONFIG_CONTIG)
		{
			/* can rely on libjpeg to convert to RGB */
			/* XXX should restore current state on exit */
			TIFFSetField(tif, TIFFTAG_JPEGCOLORMODE, JPEGCOLORMODE_RGB);
			img->photometric = PHOTOMETRIC_RGB;
		}
		break;
	case PHOTOMETRIC_RGB: 
		if (colorchannels < 3)
		{
			sprintf(emsg, "Sorry, can not handle RGB image with %s=%d",
								"Color channels", colorchannels);
			return 0;
		}
		break;
	case PHOTOMETRIC_SEPARATED:
		{
			uint16 inkset;
			TIFFGetFieldDefaulted(tif, TIFFTAG_INKSET, &inkset);
			if (inkset != INKSET_CMYK)
			{
				sprintf(emsg, "Sorry, can not handle separated image with %s=%d",
									"InkSet", inkset);
				return 0;
			}
			if (img->samplesperpixel != 4)
			{
				sprintf(emsg, "Sorry, can not handle separated image with %s=%d",
									"Samples/pixel", img->samplesperpixel);
				return 0;
			}
		}
		break;
	case PHOTOMETRIC_LOGL:
		if (compress != COMPRESSION_SGILOG)
		{
			sprintf(emsg, "Sorry, LogL data must have %s=%d",
								"Compression", COMPRESSION_SGILOG);
			return 0;
		}
		TIFFSetField(tif, TIFFTAG_SGILOGDATAFMT, SGILOGDATAFMT_8BIT);
		img->photometric = PHOTOMETRIC_MINISBLACK;	/* little white lie */
		img->bitspersample = 8;
		break;
	case PHOTOMETRIC_LOGLUV:
		if (compress != COMPRESSION_SGILOG && compress != COMPRESSION_SGILOG24)
		{
			sprintf(emsg, "Sorry, LogLuv data must have %s=%d or %d",
								"Compression", COMPRESSION_SGILOG, COMPRESSION_SGILOG24);
			return 0;
		}
		if (planarconfig != PLANARCONFIG_CONTIG)
		{
			sprintf(emsg, "Sorry, can not handle LogLuv images with %s=%d",
								"Planarconfiguration", planarconfig);
			return 0;
		}
		TIFFSetField(tif, TIFFTAG_SGILOGDATAFMT, SGILOGDATAFMT_8BIT);
		img->photometric = PHOTOMETRIC_RGB;		/* little white lie */
		img->bitspersample = 8;
		break;
	default:
		sprintf(emsg, "Sorry, can not handle image with %s=%d",
							photoTag, img->photometric);
		return 0;
	}
	img->Map = NULL;
	img->BWmap = NULL;
	img->PALmap = NULL;
	img->ycbcr = NULL;
	TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &img->width);
	TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &img->height);
	TIFFGetFieldDefaulted(tif, TIFFTAG_ORIENTATION, &img->orientation);
	img->isContig = !(planarconfig == PLANARCONFIG_SEPARATE && colorchannels > 1);
	if (img->isContig)
	{
		img->get = TIFFIsTiled(tif) ? gtTileContig : gtStripContig;
		(void) pickTileContigCase(img);
	}
	else
	{
		img->get = TIFFIsTiled(tif) ? gtTileSeparate : gtStripSeparate;
		(void) pickTileSeparateCase(img);
	}
	return 1;
}


int
TIFFPixImageGet(TIFFRGBAImage * img, char * pixBits, uint32 rowBytes, uint32 w, uint32 h)
{
	if (img->get == NULL)
	{
		TIFFError(TIFFFileName(img->tif),
					"No \"get\" routine setup");
		return 0;
	}
	if (img->put.any == NULL)
	{
		TIFFError(TIFFFileName(img->tif),
					"No \"put\" routine setup; probably can not handle image format");
		return 0;
	}
	return (*img->get)(img, (unsigned char *) pixBits, rowBytes, w, h);
}


int
TIFFRGBAImageGet(TIFFRGBAImage* img, uint32* raster, uint32 w, uint32 h)
{
	if (img->get == NULL)
	{
		TIFFError(TIFFFileName(img->tif),
					"No \"get\" routine setup");
		return 0;
	}
	if (img->put.any == NULL)
	{
		TIFFError(TIFFFileName(img->tif),
					"No \"put\" routine setup; probably can not handle image format");
		return 0;
	}
	return (*img->get)(img, (unsigned char *) raster, w, w, h);
}


/*
 * Read the specified image into an ABGR-format raster.
 */
int
TIFFReadRGBAImage(TIFF* tif, uint32 rwidth, uint32 rheight, uint32* raster, int stop)
{
	char emsg[256];
	TIFFRGBAImage img;
	int ok;

	if (TIFFRGBAImageBegin(&img, tif, stop, emsg))
	{
		/* XXX verify rwidth and rheight against width and height */
		ok = TIFFRGBAImageGet(&img, raster+(rheight-img.height)*rwidth, rwidth, img.height);
		TIFFRGBAImageEnd(&img);
	}
	else
	{
		TIFFError(TIFFFileName(tif), emsg);
		ok = 0;
	}
	return (ok);
}


static uint32
SetOrientation(TIFFRGBAImage* img, uint32 h)
{
	TIFF* tif = img->tif;
	uint32 y;

	switch (img->orientation)
	{
	case ORIENTATION_BOTRIGHT:
	case ORIENTATION_RIGHTBOT:	/* XXX */
	case ORIENTATION_LEFTBOT:	/* XXX */
		TIFFWarning(TIFFFileName(tif), "using bottom-left orientation");
		img->orientation = ORIENTATION_BOTLEFT;
		/* fall thru... */
	case ORIENTATION_BOTLEFT:
		y = 0;
		break;
	case ORIENTATION_TOPRIGHT:
	case ORIENTATION_RIGHTTOP:	/* XXX */
	case ORIENTATION_LEFTTOP:	/* XXX */
	default:
		TIFFWarning(TIFFFileName(tif), "using top-left orientation");
		img->orientation = ORIENTATION_TOPLEFT;
		/* fall thru... */
	case ORIENTATION_TOPLEFT:
		y = h - 1;
		break;
	}
	return y;
}

/*
 * Get a tile-organized image that has
 *		SamplesPerPixel == 1
 * or
 *		PlanarConfiguration contiguous if SamplesPerPixel > 1
 */	
static int
gtTileContig(TIFFRGBAImage * img, u_char * raster, uint32 rowBytes, uint32 wd, uint32 ht)
{
	TIFF *	tif = img->tif;
	uint32	col, row, y;
	uint32	clipWd, tileWd, tileHt;
	int32		fromskew, toskew;
	uint32	nrows;

	u_char * buf = (u_char *) _TIFFmalloc(TIFFTileSize(tif));
	if (buf == 0)
	{
		TIFFError(TIFFFileName(tif), "No space for tile buffer");
		return 0;
	}

	TIFFGetField(tif, TIFFTAG_TILEWIDTH, &tileWd);
	TIFFGetField(tif, TIFFTAG_TILELENGTH, &tileHt);
	y = SetOrientation(img, ht);
	toskew = -(int32) (img->orientation == ORIENTATION_TOPLEFT ? tileWd+wd : tileWd-wd);
	for (row = 0; row < ht; row += tileHt)
	{
		if ((nrows = ht - row) > tileHt)
			nrows = tileHt;
		for (col = 0; col < wd; col += tileWd)
		{
			if (TIFFReadTile(tif, buf, col, row, 0, 0) < 0 && img->stoponerr)
				break;
			if (col + tileWd > wd)
			{
				/*
				* Tile is clipped horizontally.  Calculate
				* visible portion and skewing factors.
				*/
				clipWd = wd - col;
				fromskew = tileWd - clipWd;
			}
			else
			{
				clipWd = tileWd;
				fromskew = 0;
			}
			(*(img->put.contig))(img, raster+y*wd+col, col, y, clipWd, nrows, fromskew, toskew + fromskew, buf);
		/* TO DO: Sample down from 8-bit greyscale to depth supported by Newton device. */
		}
		y += (img->orientation == ORIENTATION_TOPLEFT ? -(int32) nrows : (int32) nrows);
	}
	_TIFFfree(buf);
	return 1;
}

/*
 * Get a tile-organized image that has
 * 	SamplesPerPixel > 1
 * 	PlanarConfiguration separated
 * We assume that all such images are RGB.
 */	
static int
gtTileSeparate(TIFFRGBAImage * img, u_char * raster, uint32 rowBytes, uint32 wd, uint32 ht)
{
	TIFF *	tif = img->tif;
	uint32	col, row, y;
	uint32	clipWd, tileWd, tileHt;
	int32		fromskew, toskew;
	uint32	nrows;
	u_char *	r, * g, * b, * a;
	int		alpha = img->alpha;

	tsize_t	tilesize = TIFFTileSize(tif);
	u_char * buf = (u_char *) _TIFFmalloc(4*tilesize);
	if (buf == 0)
	{
		TIFFError(TIFFFileName(tif), "No space for tile buffer");
		return 0;
	}

	r = buf;
	g = r + tilesize;
	b = g + tilesize;
	a = b + tilesize;
	if (!alpha)
		memset(a, 0xFF, tilesize);
	TIFFGetField(tif, TIFFTAG_TILEWIDTH, &tileWd);
	TIFFGetField(tif, TIFFTAG_TILELENGTH, &tileHt);
	y = SetOrientation(img, ht);
	toskew = -(int32) (img->orientation == ORIENTATION_TOPLEFT ? tileWd+wd : tileWd-wd);
	for (row = 0; row < ht; row += tileHt)
	{
		if ((nrows = ht - row) > tileHt)
			nrows = tileHt;
		for (col = 0; col < wd; col += tileWd)
		{
			if (TIFFReadTile(tif, r, col, row,0,0) < 0 && img->stoponerr)
				break;
			if (TIFFReadTile(tif, g, col, row,0,1) < 0 && img->stoponerr)
				break;
			if (TIFFReadTile(tif, b, col, row,0,2) < 0 && img->stoponerr)
				break;
			if (alpha &&
			    TIFFReadTile(tif, a, col, row,0,3) < 0 && img->stoponerr)
				break;
			if (col + tileWd > wd)
			{
				/*
				* Tile is clipped horizontally.  Calculate
				* visible portion and skewing factors.
				*/
				clipWd = wd - col;
				fromskew = tileWd - clipWd;
			}
			else
			{
				clipWd = tileWd;
				fromskew = 0;
			}
			(*(img->put.separate))(img, raster+y*wd+col, col, y, clipWd, nrows, fromskew, toskew + fromskew, r, g, b, a);
		/* TO DO: Sample down from 8-bit greyscale to depth supported by Newton device. */
		}
		y += (img->orientation == ORIENTATION_TOPLEFT ? -(int32) nrows : (int32) nrows);
	}
	_TIFFfree(buf);
	return 1;
}


/*
 * Get a strip-organized image that has
 *		SamplesPerPixel == 1
 * or
 *		PlanarConfiguration contiguous
 */	
static int
gtStripContig(TIFFRGBAImage * img, u_char * pixBits, uint32 rowBytes, uint32 wd, uint32 ht)
{
	TIFF *	tif = img->tif;
	uint32	v, y, row;
	uint32	rowsperstrip, nrows;
	tsize_t	scanlinesize;
	int32		fromskew, toskew;

	u_char *	srcPtr, * pixBuf, * pixBufOrigin;
	u_char *	buf = (u_char *) _TIFFmalloc(TIFFStripSize(tif));
	if (buf == NULL)
	{
		TIFFError(TIFFFileName(tif), "No space for strip buffer");
		return 0;
	}

	TIFFGetFieldDefaulted(tif, TIFFTAG_ROWSPERSTRIP, &rowsperstrip);
	pixBuf = (u_char *) _TIFFmalloc(rowsperstrip * wd);
	if (pixBuf == NULL)
	{
		TIFFError(TIFFFileName(tif), "No space for greyscale buffer");
		return 0;
	}

	scanlinesize = TIFFScanlineSize(tif);
	y = SetOrientation(img, ht);
	toskew = -(int32) (img->orientation == ORIENTATION_TOPLEFT ? wd-wd : wd+wd);
	if ((fromskew = img->width - wd) < 0)
		fromskew = 0;
	for (row = 0; row < ht; row += rowsperstrip)
	{
		if ((nrows = ht - row) > rowsperstrip)
			nrows = rowsperstrip;
		if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 0), buf, nrows*scanlinesize) < 0 && img->stoponerr)
			break;
		pixBufOrigin = (img->orientation == ORIENTATION_TOPLEFT) ? pixBuf : pixBuf + (nrows - 1) * wd;
		(*(img->put.contig))(img, pixBufOrigin, 0, y, wd, nrows, fromskew, toskew, buf);

		/* Sample down from 8-bit greyscale to depth supported by Newton device. */
		srcPtr = pixBuf;
		for (v = 0; v < nrows; v++)
		{
			u_char * dstPtr = (u_char *) pixBits + rowBytes * (row + v);
			u_char pixHi, pixLo;
			unsigned i;
			for (i = wd/2; i > 0; i--)
			{
				pixHi = *srcPtr++ & 0xF0;
				pixLo = *srcPtr++ >> 4;
				*dstPtr++ = pixHi | pixLo;
			}
			if (wd & 1)	// if width is odd, write out odd pixel
				*dstPtr = *srcPtr++ & 0xF0;
		}

		y += (img->orientation == ORIENTATION_TOPLEFT ? (int32) nrows : (int32) -nrows);
	}
	_TIFFfree(pixBuf);
	_TIFFfree(buf);
	return 1;
}


/*
 * Get a strip-organized image with
 *	 SamplesPerPixel > 1
 *	 PlanarConfiguration separated
 * We assume that all such images are RGB.
 */
static int
gtStripSeparate(TIFFRGBAImage * img, u_char * pixBits, uint32 rowBytes, uint32 wd, uint32 ht)
{
	TIFF *	tif = img->tif;
	uint32	v, y, row;
	uint32	rowsperstrip, nrows;
	uint32	imagewidth = img->width;
	int32		fromskew, toskew;
	u_char *	r, * g, * b, * a;
	int		alpha = img->alpha;
	tsize_t	scanlinesize;
	tsize_t	stripsize = TIFFStripSize(tif);

	u_char *	srcPtr, * pixBuf, * pixBufOrigin;
	u_char * buf = (u_char *) _TIFFmalloc(4*stripsize);
	if (buf == 0)
	{
		TIFFError(TIFFFileName(tif), "No space for tile buffer");
		return 0;
	}

	r = buf;
	g = r + stripsize;
	b = g + stripsize;
	a = b + stripsize;
	if (!alpha)
		memset(a, 0xFF, stripsize);
	TIFFGetFieldDefaulted(tif, TIFFTAG_ROWSPERSTRIP, &rowsperstrip);
	scanlinesize = TIFFScanlineSize(tif);
	y = SetOrientation(img, ht);
	toskew = -(int32) (img->orientation == ORIENTATION_TOPLEFT ? wd+wd : wd-wd);
	if ((fromskew = imagewidth - wd) < 0)
		fromskew = 0;
	for (row = 0; row < ht; row += rowsperstrip)
	{
		if ((nrows = ht - row) > rowsperstrip)
			nrows = rowsperstrip;
		if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 0), r, nrows*scanlinesize) < 0 && img->stoponerr)
			break;
		if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 1), g, nrows*scanlinesize) < 0 && img->stoponerr)
			break;
		if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 2), b, nrows*scanlinesize) < 0 && img->stoponerr)
			break;
		if (alpha &&
			 TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 3), a, nrows*scanlinesize) < 0 && img->stoponerr)
			break;
		pixBufOrigin = (img->orientation == ORIENTATION_TOPLEFT) ? pixBuf + (nrows - 1) * wd : pixBuf;
		(*(img->put.separate))(img, pixBufOrigin, 0, y, wd, nrows, fromskew, toskew, r, g, b, a);

		/* Sample down from 8-bit greyscale to depth supported by Newton device. */
		srcPtr = pixBuf;
		for (v = 0; v < nrows; v++)
		{
			u_char * dstPtr = (u_char *) pixBits + rowBytes * (row + v);
			u_char pixHi, pixLo;
			unsigned i;
			for (i = wd/2; i > 0; i--)
			{
				pixHi = *srcPtr++ & 0xF0;
				pixLo = *srcPtr++ >> 4;
				*dstPtr++ = pixHi | pixLo;
			}
			if (wd & 1)	// if width is odd, write out odd pixel
				*dstPtr = *srcPtr++ & 0xF0;
		}

		y += (img->orientation == ORIENTATION_TOPLEFT ? -(int32) nrows : (int32) nrows);
	}
	_TIFFfree(buf);
	return 1;
}


/*
 * The following routines move decoded data returned
 * from the TIFF library into rasters filled with packed
 * 8-bit greyscale pixels (ie suitable for a Newton bitmap
 * when sampled down to 4 bits).
 *
 * The routines have been created according to the most
 * important cases and optimized.  pickTileContigCase and
 * pickTileSeparateCase analyze the parameters and select
 * the appropriate "put" routine to use.
 */
#define	REPEAT2(op)	op; op
#define	REPEAT4(op)	op; op; op; op
#define	REPEAT8(op)	op; op; op; op; op; op; op; op
#define	CASE4(x,op)	switch (x) { case 3: op; case 2: op; case 1: op; }
#define	CASE8(x,op) switch (x) { case 7: op; case 6: op; case 5: op; case 4: op; case 3: op; case 2: op; case 1: op; }
#define	NOP

#define	UNROLL2(w, op1, op2) {		\
	uint32 _x;			\
	for (_x = w; _x >= 2; _x -= 2) {	\
		op1;				\
		REPEAT2(op2);	\
	}						\
	if (_x) {			\
		op1;				\
		op2;				\
	}	\
}
#define	UNROLL4(w, op1, op2) {		\
	uint32 _x;			\
	for (_x = w; _x >= 4; _x -= 4) {	\
		op1;				\
		REPEAT4(op2);	\
	}						\
	if (_x > 0) {		\
		op1;				\
		CASE4(_x,op2);	\
	}	\
}
#define	UNROLL8(w, op1, op2) {		\
	uint32 _x;			\
	for (_x = w; _x >= 8; _x -= 8) {	\
		op1;				\
		REPEAT8(op2);	\
	}						\
	if (_x > 0) {		\
		op1;				\
		CASE8(_x,op2);	\
	}	\
}

#define SKEW(r,g,b,skew)		{ r += skew; g += skew; b += skew; }
#define SKEW4(r,g,b,a,skew)	{ r += skew; g += skew; b += skew; a+= skew; }

#define A1 ((uint32)(0xffL<<24))
#define PACK(r,g,b)		((uint32)(r)|((uint32)(g)<<8)|((uint32)(b)<<16)|A1)
#define PACK4(r,g,b,a)	((uint32)(r)|((uint32)(g)<<8)|((uint32)(b)<<16)|((uint32)(a)<<24))
#define W2B(v) (((v)>>8)&0xff)
#define PACKW(r,g,b)		((uint32)W2B(r)|((uint32)W2B(g)<<8)|((uint32)W2B(b)<<16)|A1)
#define PACKW4(r,g,b,a)	((uint32)W2B(r)|((uint32)W2B(g)<<8)|((uint32)W2B(b)<<16)|((uint32)W2B(a)<<24))

#define	DECLAREContigPutFunc(name) \
static void name(	TIFFRGBAImage * img, \
						u_char * pixPtr, \
						uint32 x, uint32 y, uint32 wd, uint32 ht, \
						int32 fromskew, int32 toskew, \
						u_char * srcPtr )

/*
 * -------- P A L E T T E --------
 *
 * 8-bit palette => greyscale
 */
DECLAREContigPutFunc(put8bitcmaptile)
{
	u_char ** colorMap = img->PALmap;

	while (ht-- > 0)
	{
		UNROLL8(wd, NOP, *pixPtr++ = colorMap[*srcPtr++][0]);
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * 4-bit palette => greyscale
 */
DECLAREContigPutFunc(put4bitcmaptile)
{
	u_char ** colorMap = img->PALmap;
	u_char * bw;
	
	fromskew /= 2;
	while (ht-- > 0)
	{
		UNROLL2(wd, bw = colorMap[*srcPtr++], *pixPtr++ = *bw++);
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * 2-bit palette => greyscale
 */
DECLAREContigPutFunc(put2bitcmaptile)
{
	u_char ** colorMap = img->PALmap;
	u_char * bw;

	fromskew /= 4;
	while (ht-- > 0)
	{
		UNROLL4(wd, bw = colorMap[*srcPtr++], *pixPtr++ = *bw++);
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * 1-bit palette => greyscale
 */
DECLAREContigPutFunc(put1bitcmaptile)
{
	u_char ** colorMap = img->PALmap;
	u_char * bw;
	
	fromskew /= 8;
	while (ht-- > 0)
	{
		UNROLL8(wd, bw = colorMap[*srcPtr++], *pixPtr++ = *bw++);
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * -------- G R E Y S C A L E --------
 *
 * 8-bit greyscale => greyscale
 */
DECLAREContigPutFunc(putgreytile)
{
	u_char ** bwMap = img->BWmap;

	while (ht-- > 0)
	{
		for (x = wd; x-- > 0; )
			*pixPtr++ = bwMap[*srcPtr++][0];
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * 4-bit greyscale => greyscale
 */
DECLAREContigPutFunc(put4bitbwtile)
{
	u_char ** bwMap = img->BWmap;
	u_char * bw;

	fromskew /= 2;
	while (ht-- > 0)
	{
		UNROLL2(wd, bw = bwMap[*srcPtr++], *pixPtr++ = *bw++);
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * 2-bit greyscale => greyscale
 */
DECLAREContigPutFunc(put2bitbwtile)
{
	u_char ** bwMap = img->BWmap;
	u_char * bw;

	fromskew /= 4;
	while (ht-- > 0)
	{
		UNROLL4(wd, bw = bwMap[*srcPtr++], *pixPtr++ = *bw++);
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * 1-bit bilevel => greyscale
 */
DECLAREContigPutFunc(put1bitbwtile)
{
	u_char ** bwMap = img->BWmap;
	u_char * bw;

	fromskew /= 8;
	while (ht-- > 0)
	{
		UNROLL8(wd, bw = bwMap[*srcPtr++], *pixPtr++ = *bw++);
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * -------- P A C K E D   S A M P L E S --------
 *
 * 8-bit packed samples, w/o map => greyscale
 */
DECLAREContigPutFunc(putRGBcontig8bittile)
{
	int samplesperpixel = img->samplesperpixel;

	fromskew *= samplesperpixel;
	while (ht-- > 0)
	{
		UNROLL8(wd, NOP, *pixPtr++ = MapRGB(srcPtr[0], srcPtr[1], srcPtr[2]); srcPtr += samplesperpixel);
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * 8-bit packed samples, w/ map => greyscale
 */
DECLAREContigPutFunc(putRGBcontig8bitMaptile)
{
	TIFFRGBValue * map = img->Map;
	int samplesperpixel = img->samplesperpixel;

	fromskew *= samplesperpixel;
	while (ht-- > 0)
	{
		for (x = wd; x-- > 0;)
		{
			*pixPtr++ = MapRGB(map[srcPtr[0]], map[srcPtr[1]], map[srcPtr[2]]);
			srcPtr += samplesperpixel;
		}
		srcPtr += fromskew;
		pixPtr += toskew;
	}
}

/*
 * 8-bit packed samples => greyscale w/ associated alpha
 * (known to have map == NULL)
 */
DECLAREContigPutFunc(putRGBAAcontig8bittile)
{
	int samplesperpixel = img->samplesperpixel;

	fromskew *= samplesperpixel;
	while (ht-- > 0)
	{
		UNROLL8(wd, NOP,	*pixPtr++ = MapRGBA(srcPtr[0], srcPtr[1], srcPtr[2], srcPtr[3]);
								srcPtr += samplesperpixel);
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * 8-bit packed samples => greyscale w/ unassociated alpha
 * (known to have map == NULL)
 */
DECLAREContigPutFunc(putRGBUAcontig8bittile)
{
	int samplesperpixel = img->samplesperpixel;

	fromskew *= samplesperpixel;
	while (ht-- > 0)
	{
		for (x = wd; x-- > 0;)
		{
			*pixPtr++ = MapRGBA(srcPtr[0], srcPtr[1], srcPtr[2], srcPtr[3]);
			srcPtr += samplesperpixel;
		}
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * -------- P A C K E D   1 6 - B I T   S A M P L E S --------
 *
 * 16-bit packed samples => greyscale
 */
DECLAREContigPutFunc(putRGBcontig16bittile)
{
	int samplesperpixel = img->samplesperpixel;
	uint16 * wPtr = (uint16 *) srcPtr;

	fromskew *= samplesperpixel;
	while (ht-- > 0)
	{
		for (x = wd; x-- > 0;)
		{
			*pixPtr++ = MapRGB(wPtr[0], wPtr[1], wPtr[2]);
			wPtr += samplesperpixel;
		}
		pixPtr += toskew;
		wPtr += fromskew;
	}
}

/*
 * 16-bit packed samples => greyscale w/ associated alpha
 * (known to have map == NULL)
 */
DECLAREContigPutFunc(putRGBAAcontig16bittile)
{
	int samplesperpixel = img->samplesperpixel;
	uint16 * wPtr = (uint16 *) srcPtr;

	fromskew *= samplesperpixel;
	while (ht-- > 0)
	{
		for (x = wd; x-- > 0;)
		{
			*pixPtr++ = MapRGBA(wPtr[0], wPtr[1], wPtr[2], wPtr[3]);
			wPtr += samplesperpixel;
		}
		pixPtr += toskew;
		wPtr += fromskew;
	}
}

/*
 * 16-bit packed samples => greyscale w/ unassociated alpha
 * (known to have map == NULL)
 */
DECLAREContigPutFunc(putRGBUAcontig16bittile)
{
	int samplesperpixel = img->samplesperpixel;
	uint16 * wPtr = (uint16 *) srcPtr;

	fromskew *= samplesperpixel;
	while (ht-- > 0)
	{
		uint32 r,g,b,a;
		/*
			* We shift alpha down four bits just in case unsigned
			* arithmetic doesn't handle the full range.
			* We still have plenty of accuracy, since the output is 8 bits.
			* So we have (r * 0xffff) * (a * 0xfff)) = r*a * (0xffff*0xfff)
			* Since we want r*a * 0xff for eight bit output,
			* we divide by (0xffff * 0xfff) / 0xff == 0x10eff.
			*/
		for (x = wd; x-- > 0;)
		{
			a = wPtr[3] >> 4; 
			r = (wPtr[0] * a) / 0x10eff;
			g = (wPtr[1] * a) / 0x10eff;
			b = (wPtr[2] * a) / 0x10eff;
			*pixPtr++ = MapRGB(r,g,b);
			wPtr += samplesperpixel;
		}
		pixPtr += toskew;
		wPtr += fromskew;
	}
}

/*
 * -------- P A C K E D   C M Y K   S A M P L E S --------
 *
 * 8-bit packed CMYK samples w/o map => greyscale
 *
 * NB: The conversion of CMYK->RGB is *very* crude.
 */
DECLAREContigPutFunc(putRGBcontig8bitCMYKtile)
{
	int samplesperpixel = img->samplesperpixel;
	uint16 r, g, b, k;

	fromskew *= samplesperpixel;
	while (ht-- > 0)
	{
		UNROLL8(wd, NOP,	k = 255 - srcPtr[3];
								r = (k*(255-srcPtr[0]))/255;
								g = (k*(255-srcPtr[1]))/255;
								b = (k*(255-srcPtr[2]))/255;
								*pixPtr++ = MapRGB(r, g, b);
								srcPtr += samplesperpixel);
		pixPtr += toskew;
		srcPtr += fromskew;
	}
}

/*
 * 8-bit packed CMYK samples w/ map => greyscale
 *
 * NB: The conversion of CMYK->RGB is *very* crude.
 */
DECLAREContigPutFunc(putRGBcontig8bitCMYKMaptile)
{
	int samplesperpixel = img->samplesperpixel;
	TIFFRGBValue * map = img->Map;
	uint16 r, g, b, k;

	fromskew *= samplesperpixel;
	while (ht-- > 0)
	{
		for (x = wd; x-- > 0;)
		{
			k = 255 - srcPtr[3];
			r = (k*(255-srcPtr[0]))/255;
			g = (k*(255-srcPtr[1]))/255;
			b = (k*(255-srcPtr[2]))/255;
			*pixPtr++ = MapRGB(map[r], map[g], map[b]);
			srcPtr += samplesperpixel;
		}
		srcPtr += fromskew;
		pixPtr += toskew;
	}
}

#define	DECLARESepPutFunc(name) \
static void name(	TIFFRGBAImage * img,\
						u_char * pixPtr,\
						uint32 x, uint32 y, uint32 wd, uint32 ht,\
						int32 fromskew, int32 toskew,\
						u_char * r, u_char * g, u_char * b, u_char * a )

/*
 * -------- U N P A C K E D   S A M P L E S --------
 *
 * 8-bit unpacked samples => greyscale
 */
DECLARESepPutFunc(putRGBseparate8bittile)
{
	while (ht-- > 0)
	{
		UNROLL8(wd, NOP, *pixPtr++ = MapRGB(*r++, *g++, *b++));
		SKEW(r, g, b, fromskew);
		pixPtr += toskew;
	}
}

/*
 * 8-bit unpacked samples w/ map => greyscale
 */
DECLARESepPutFunc(putRGBseparate8bitMaptile)
{
	TIFFRGBValue * map = img->Map;

	while (ht-- > 0)
	{
		for (x = wd; x > 0; x--)
			*pixPtr++ = MapRGB(map[*r++], map[*g++], map[*b++]);
		SKEW(r, g, b, fromskew);
		pixPtr += toskew;
	}
}

/*
 * 8-bit unpacked samples => greyscale w/ associated alpha
 */
DECLARESepPutFunc(putRGBAAseparate8bittile)
{
	while (ht-- > 0)
	{
		UNROLL8(wd, NOP, *pixPtr++ = MapRGBA(*r++, *g++, *b++, *a++));
		SKEW4(r, g, b, a, fromskew);
		pixPtr += toskew;
	}
}

/*
 * 8-bit unpacked samples => greyscale w/ unassociated alpha
 */
DECLARESepPutFunc(putRGBUAseparate8bittile)
{
	while (ht-- > 0)
	{
		for (x = wd; x-- > 0;)
		{
			*pixPtr++ = MapRGBA(*r++, *g++, *b++, *a++);
		}
		SKEW4(r, g, b, a, fromskew);
		pixPtr += toskew;
	}
}

/*
 * -------- U N P A C K E D   1 6 - B I T   S A M P L E S --------
 *
 * 16-bit unpacked samples => greyscale
 */
DECLARESepPutFunc(putRGBseparate16bittile)
{
	uint16 *wr = (uint16*) r;
	uint16 *wg = (uint16*) g;
	uint16 *wb = (uint16*) b;

	while (ht-- > 0)
	{
		for (x = 0; x < wd; x++)
			*pixPtr++ = MapRGB(*wr++, *wg++, *wb++);
		SKEW(wr, wg, wb, fromskew);
		pixPtr += toskew;
	}
}

/*
 * 16-bit unpacked samples => greyscale w/ associated alpha
 */
DECLARESepPutFunc(putRGBAAseparate16bittile)
{
	uint16 *wr = (uint16*) r;
	uint16 *wg = (uint16*) g;
	uint16 *wb = (uint16*) b;
	uint16 *wa = (uint16*) a;

	while (ht-- > 0)
	{
		for (x = 0; x < wd; x++)
			*pixPtr++ = MapRGBA(*wr++, *wg++, *wb++, *wa++);
		SKEW4(wr, wg, wb, wa, fromskew);
		pixPtr += toskew;
	}
}

/*
 * 16-bit unpacked samples => greyscale w/ unassociated alpha
 */
DECLARESepPutFunc(putRGBUAseparate16bittile)
{
	uint16 *wr = (uint16*) r;
	uint16 *wg = (uint16*) g;
	uint16 *wb = (uint16*) b;
	uint16 *wa = (uint16*) a;

	while (ht-- > 0)
	{
		uint32 r,g,b,a;
		/*
			* We shift alpha down four bits just in case unsigned
			* arithmetic doesn't handle the full range.
			* We still have plenty of accuracy, since the output is 8 bits.
			* So we have (r * 0xffff) * (a * 0xfff)) = r*a * (0xffff*0xfff)
			* Since we want r*a * 0xff for eight bit output,
			* we divide by (0xffff * 0xfff) / 0xff == 0x10eff.
			*/
		for (x = wd; x-- > 0;)
		{
			a = *wa++ >> 4; 
			r = (*wr++ * a) / 0x10eff;
			g = (*wg++ * a) / 0x10eff;
			b = (*wb++ * a) / 0x10eff;
			*pixPtr++ = MapRGBA(r,g,b,a);
		}
		SKEW4(wr, wg, wb, wa, fromskew);
		pixPtr += toskew;
	}
}


#if defined(YCBCR_SUPPORTED)
/*
 * YCbCr -> RGB conversion and packing routines.  The colorspace
 * conversion algorithm comes from the IJG v5a code; see below
 * for more information on how it works.
 */

#define	YCbCrtoRGB(dst, yc) {					\
	int Y = (yc);										\
	dst = PACK(											\
		clamptab[Y+Crrtab[Cr]],						\
		clamptab[Y + (int)((Cbgtab[Cb]+Crgtab[Cr])>>16)],	\
		clamptab[Y+Cbbtab[Cb]]);					\
}
#define	YCbCrSetup									\
	TIFFYCbCrToRGB* ycbcr = img->ycbcr;			\
	int* Crrtab = ycbcr->Cr_r_tab;				\
	int* Cbbtab = ycbcr->Cb_b_tab;				\
	int32* Crgtab = ycbcr->Cr_g_tab;				\
	int32* Cbgtab = ycbcr->Cb_g_tab;				\
	TIFFRGBValue* clamptab = ycbcr->clamptab

/*
 * 8-bit packed YCbCr samples w/ 4,4 subsampling => RGB
 */
DECLAREContigPutFunc(putcontig8bitYCbCr44tile)
{
	YCbCrSetup;
	u_char * cp1 = pixPtr+wd+toskew;
	u_char * cp2 =    cp1+wd+toskew;
	u_char * cp3 =    cp2+wd+toskew;
	int32   incr = 3*wd + 4*toskew;

	(void) y;
	/* XXX adjust fromskew */
	for (; ht >= 4; ht -= 4)
	{
		x = wd>>2;
		do {
			int Cb = srcPtr[16];
			int Cr = srcPtr[17];

			YCbCrtoRGB(pixPtr [0], srcPtr[ 0]);
			YCbCrtoRGB(pixPtr [1], srcPtr[ 1]);
			YCbCrtoRGB(pixPtr [2], srcPtr[ 2]);
			YCbCrtoRGB(pixPtr [3], srcPtr[ 3]);
			YCbCrtoRGB(cp1[0], srcPtr[ 4]);
			YCbCrtoRGB(cp1[1], srcPtr[ 5]);
			YCbCrtoRGB(cp1[2], srcPtr[ 6]);
			YCbCrtoRGB(cp1[3], srcPtr[ 7]);
			YCbCrtoRGB(cp2[0], srcPtr[ 8]);
			YCbCrtoRGB(cp2[1], srcPtr[ 9]);
			YCbCrtoRGB(cp2[2], srcPtr[10]);
			YCbCrtoRGB(cp2[3], srcPtr[11]);
			YCbCrtoRGB(cp3[0], srcPtr[12]);
			YCbCrtoRGB(cp3[1], srcPtr[13]);
			YCbCrtoRGB(cp3[2], srcPtr[14]);
			YCbCrtoRGB(cp3[3], srcPtr[15]);

			pixPtr += 4, cp1 += 4, cp2 += 4, cp3 += 4;
			srcPtr += 18;
		} while (--x);
		pixPtr += incr, cp1 += incr, cp2 += incr, cp3 += incr;
		srcPtr += fromskew;
	}
}

/*
 * 8-bit packed YCbCr samples w/ 4,2 subsampling => RGB
 */
DECLAREContigPutFunc(putcontig8bitYCbCr42tile)
{
	YCbCrSetup;
	u_char * cp1 = pixPtr+wd+toskew;
	int32   incr = 2*toskew+wd;

	(void) y;
	/* XXX adjust fromskew */
	for (; ht >= 2; ht -= 2)
	{
		x = wd>>2;
		do {
			int Cb = srcPtr[8];
			int Cr = srcPtr[9];
	
			YCbCrtoRGB(pixPtr [0], srcPtr[0]);
			YCbCrtoRGB(pixPtr [1], srcPtr[1]);
			YCbCrtoRGB(pixPtr [2], srcPtr[2]);
			YCbCrtoRGB(pixPtr [3], srcPtr[3]);
			YCbCrtoRGB(cp1[0], srcPtr[4]);
			YCbCrtoRGB(cp1[1], srcPtr[5]);
			YCbCrtoRGB(cp1[2], srcPtr[6]);
			YCbCrtoRGB(cp1[3], srcPtr[7]);
	
			pixPtr += 4, cp1 += 4;
			srcPtr += 10;
		} while (--x);
		pixPtr += incr, cp1 += incr;
		srcPtr += fromskew;
	}
}

/*
 * 8-bit packed YCbCr samples w/ 4,1 subsampling => RGB
 */
DECLAREContigPutFunc(putcontig8bitYCbCr41tile)
{
	YCbCrSetup;

	(void) y;
	/* XXX adjust fromskew */
	do {
		x = wd>>2;
		do {
			int Cb = srcPtr[4];
			int Cr = srcPtr[5];
	
			YCbCrtoRGB(pixPtr [0], srcPtr[0]);
			YCbCrtoRGB(pixPtr [1], srcPtr[1]);
			YCbCrtoRGB(pixPtr [2], srcPtr[2]);
			YCbCrtoRGB(pixPtr [3], srcPtr[3]);
	
			pixPtr += 4;
			srcPtr += 6;
		} while (--x);
		pixPtr += toskew;
		srcPtr += fromskew;
	} while (--ht);
}

/*
 * 8-bit packed YCbCr samples w/ 2,2 subsampling => RGB
 */
DECLAREContigPutFunc(putcontig8bitYCbCr22tile)
{
	YCbCrSetup;
	u_char * cp1 = pixPtr+wd+toskew;
	int32   incr = 2*toskew+wd;

	(void) y;
	/* XXX adjust fromskew */
	for (; ht >= 2; ht -= 2)
	{
		x = wd>>1;
		do {
			int Cb = srcPtr[4];
			int Cr = srcPtr[5];

			YCbCrtoRGB(pixPtr [0], srcPtr[0]);
			YCbCrtoRGB(pixPtr [1], srcPtr[1]);
			YCbCrtoRGB(cp1[0], srcPtr[2]);
			YCbCrtoRGB(cp1[1], srcPtr[3]);
	
			pixPtr += 2, cp1 += 2;
			srcPtr += 6;
		} while (--x);
		pixPtr += incr, cp1 += incr;
		srcPtr += fromskew;
	}
}

/*
 * 8-bit packed YCbCr samples w/ 2,1 subsampling => RGB
 */
DECLAREContigPutFunc(putcontig8bitYCbCr21tile)
{
	YCbCrSetup;

	(void) y;
	/* XXX adjust fromskew */
	do {
		x = wd>>1;
		do {
			int Cb = srcPtr[2];
			int Cr = srcPtr[3];

			YCbCrtoRGB(pixPtr[0], srcPtr[0]);
			YCbCrtoRGB(pixPtr[1], srcPtr[1]);

			pixPtr += 2;
			srcPtr += 4;
		} while (--x);
		pixPtr += toskew;
		srcPtr += fromskew;
	} while (--ht);
}

/*
 * 8-bit packed YCbCr samples w/ no subsampling => RGB
 */
DECLAREContigPutFunc(putcontig8bitYCbCr11tile)
{
	YCbCrSetup;

	(void) y;
	/* XXX adjust fromskew */
	do {
		x = wd>>1;
		do {
			int Cb = srcPtr[1];
			int Cr = srcPtr[2];

			YCbCrtoRGB(*pixPtr++, srcPtr[0]);

			srcPtr += 3;
		} while (--x);
		pixPtr += toskew;
		srcPtr += fromskew;
	} while (--ht);
}

#undef	YCbCrSetup
#undef	YCbCrtoRGB

#define	LumaRed		coeffs[0]
#define	LumaGreen	coeffs[1]
#define	LumaBlue		coeffs[2]
#define	SHIFT			16
#define	FIX(x)		((int32)((x) * (1L<<SHIFT) + 0.5))
#define	ONE_HALF		((int32)(1<<(SHIFT-1)))

/*
 * Initialize the YCbCr->RGB conversion tables.  The conversion
 * is done according to the 6.0 spec:
 *
 *    R = Y + Cr*(2 - 2*LumaRed)
 *    B = Y + Cb*(2 - 2*LumaBlue)
 *    G =   Y
 *        - LumaBlue*Cb*(2-2*LumaBlue)/LumaGreen
 *        - LumaRed*Cr*(2-2*LumaRed)/LumaGreen
 *
 * To avoid floating point arithmetic the fractional constants that
 * come out of the equations are represented as fixed point values
 * in the range 0...2^16.  We also eliminate multiplications by
 * pre-calculating possible values indexed by Cb and Cr (this code
 * assumes conversion is being done for 8-bit samples).
 */
static void
TIFFYCbCrToRGBInit(TIFFYCbCrToRGB* ycbcr, TIFF* tif)
{
	TIFFRGBValue* clamptab;
	float* coeffs;
	int i;

	clamptab = (TIFFRGBValue*) ((tidata_t) ycbcr+TIFFroundup(sizeof (TIFFYCbCrToRGB), sizeof (long)));
	_TIFFmemset(clamptab, 0, 256);		/* v < 0 => 0 */
	ycbcr->clamptab = (clamptab += 256);
	for (i = 0; i < 256; i++)
		clamptab[i] = i;
	_TIFFmemset(clamptab+256, 255, 2*256);	/* v > 255 => 255 */
	TIFFGetFieldDefaulted(tif, TIFFTAG_YCBCRCOEFFICIENTS, &coeffs);
	_TIFFmemcpy(ycbcr->coeffs, coeffs, 3*sizeof (float));
	{
		float f1 = 2-2*LumaRed;					int32 D1 = FIX(f1);
		float f2 = LumaRed*f1/LumaGreen;		int32 D2 = -FIX(f2);
		float f3 = 2-2*LumaBlue;				int32 D3 = FIX(f3);
		float f4 = LumaBlue*f3/LumaGreen;	int32 D4 = -FIX(f4);
		int x;

		ycbcr->Cr_r_tab = (int*) (clamptab + 3*256);
		ycbcr->Cb_b_tab = ycbcr->Cr_r_tab + 256;
		ycbcr->Cr_g_tab = (int32*) (ycbcr->Cb_b_tab + 256);
		ycbcr->Cb_g_tab = ycbcr->Cr_g_tab + 256;
		/*
			* i is the actual input pixel value in the range 0..255
			* Cb and Cr values are in the range -128..127 (actually
			* they are in a range defined by the ReferenceBlackWhite
			* tag) so there is some range shifting to do here when
			* constructing tables indexed by the raw pixel data.
			*
			* XXX handle ReferenceBlackWhite correctly to calculate
			*     Cb/Cr values to use in constructing the tables.
			*/
		for (i = 0, x = -128; i < 256; i++, x++)
		{
			ycbcr->Cr_r_tab[i] = (int)((D1*x + ONE_HALF)>>SHIFT);
			ycbcr->Cb_b_tab[i] = (int)((D3*x + ONE_HALF)>>SHIFT);
			ycbcr->Cr_g_tab[i] = D2*x;
			ycbcr->Cb_g_tab[i] = D4*x + ONE_HALF;
		}
	}
}
#undef	SHIFT
#undef	ONE_HALF
#undef	FIX
#undef	LumaBlue
#undef	LumaGreen
#undef	LumaRed

static tileContigRoutine
initYCbCrConversion(TIFFRGBAImage* img)
{
	uint16 hs, vs;
	
	if (img->ycbcr == NULL)
	{
		img->ycbcr = (TIFFYCbCrToRGB*) _TIFFmalloc(	TIFFroundup(sizeof (TIFFYCbCrToRGB),
																	sizeof (long))
																	+ 4*256*sizeof (TIFFRGBValue)
																	+ 2*256*sizeof (int)
																	+ 2*256*sizeof (int32)	);
		if (img->ycbcr == NULL)
		{
			TIFFError(TIFFFileName(img->tif), "No space for YCbCr->RGB conversion state");
			return (NULL);
		}
		TIFFYCbCrToRGBInit(img->ycbcr, img->tif);
	}
	else
	{
		float* coeffs;

		TIFFGetFieldDefaulted(img->tif, TIFFTAG_YCBCRCOEFFICIENTS, &coeffs);
		if (_TIFFmemcmp(coeffs, img->ycbcr->coeffs, 3*sizeof (float)) != 0)
			TIFFYCbCrToRGBInit(img->ycbcr, img->tif);
	}
	/*
	* The 6.0 spec says that subsampling must be
	* one of 1, 2, or 4, and that vertical subsampling
	* must always be <= horizontal subsampling; so
	* there are only a few possibilities and we just
	* enumerate the cases.
	*/
	TIFFGetFieldDefaulted(img->tif, TIFFTAG_YCBCRSUBSAMPLING, &hs, &vs);
	switch ((hs<<4)|vs)
	{
	case 0x44: return (putcontig8bitYCbCr44tile);
	case 0x42: return (putcontig8bitYCbCr42tile);
	case 0x41: return (putcontig8bitYCbCr41tile);
	case 0x22: return (putcontig8bitYCbCr22tile);
	case 0x21: return (putcontig8bitYCbCr21tile);
	case 0x11: return (putcontig8bitYCbCr11tile);
	}
	return (NULL);
}
#endif


/*
 * Greyscale images with less than 8 bits/sample are handled
 * with a table to avoid lots of shifts and masks.  The table
 * is setup so that put*bwtile (below) can retrieve 8/bitspersample
 * pixel values simply by indexing into the table with one
 * number.
 */
static int
makebwmap(TIFFRGBAImage* img)
{
	TIFFRGBValue * map = img->Map;
	int bitspersample = img->bitspersample;
	int nsamples = 8 / bitspersample;
	u_char * p;
	unsigned i;

	img->BWmap = (u_char**) _TIFFmalloc(256*sizeof(u_char *) + 256*nsamples*sizeof(u_char));
	if (img->BWmap == NULL)
	{
		TIFFError(TIFFFileName(img->tif), "No space for B&W mapping table");
		return 0;
	}
	p = (u_char *) (img->BWmap + 256);
	for (i = 0; i < 256; i++)
	{
		img->BWmap[i] = p;
		switch (bitspersample)
		{
#define	GREY(c)	*p++ = map[c];
		case 1:
			GREY(i>>7);
			GREY((i>>6)&0x01);
			GREY((i>>5)&0x01);
			GREY((i>>4)&0x01);
			GREY((i>>3)&0x01);
			GREY((i>>2)&0x01);
			GREY((i>>1)&0x01);
			GREY(i&0x01);
			break;
		case 2:
			GREY(i>>6);
			GREY((i>>4)&0x03);
			GREY((i>>2)&0x03);
			GREY(i&0x03);
			break;
		case 4:
			GREY(i>>4);
			GREY(i&0x0F);
			break;
		case 8:
			GREY(i);
			break;
		}
#undef	GREY
	}
	return 1;
}

/*
 * Construct a mapping table to convert from the range
 * of the data samples to [0,255] --for display.  This
 * process also handles inverting B&W images when needed.
 */ 
static int
setupMap(TIFFRGBAImage* img)
{
	int32 x, range;

	range = (int32) ((1L<<img->bitspersample) - 1);
	img->Map = (TIFFRGBValue *) _TIFFmalloc((range+1) * sizeof(TIFFRGBValue));
	if (img->Map == NULL)
	{
		TIFFError(TIFFFileName(img->tif), "No space for photometric conversion table");
		return 0;
	}
	if (img->photometric == PHOTOMETRIC_MINISWHITE)
	{
		for (x = 0; x <= range; x++)
			img->Map[x] = ((range - x) * 255) / range;
	}
	else
	{
		for (x = 0; x <= range; x++)
			img->Map[x] = (x * 255) / range;
	}
	if (img->bitspersample <= 8
	 && (img->photometric == PHOTOMETRIC_MINISBLACK
	  || img->photometric == PHOTOMETRIC_MINISWHITE))
	{
		/*
		* Use photometric mapping table to construct
		* unpacking tables for samples <= 8 bits.
		*/
		if (!makebwmap(img))
			return 0;
		/* no longer need Map, free it */
		_TIFFfree(img->Map), img->Map = NULL;
	}
	return 1;
}

static int
checkcmap(TIFFRGBAImage* img)
{
	uint16* r = img->redcmap;
	uint16* g = img->greencmap;
	uint16* b = img->bluecmap;
	long n = 1L<<img->bitspersample;

	while (n-- > 0)
		if (*r++ >= 256 || *g++ >= 256 || *b++ >= 256)
			return (16);
	return (8);
}

static void
cvtcmap(TIFFRGBAImage* img)
{
	uint16* r = img->redcmap;
	uint16* g = img->greencmap;
	uint16* b = img->bluecmap;
	long i;

	for (i = (1L<<img->bitspersample) - 1; i >= 0; i--)
	{
#define	CVT(x)	((uint16)((x)>>8))
		r[i] = CVT(r[i]);
		g[i] = CVT(g[i]);
		b[i] = CVT(b[i]);
#undef	CVT
	}
}

/*
 * Palette images with <= 8 bits/sample are handled
 * with a table to avoid lots of shifts and masks.  The table
 * is setup so that put*cmaptile (below) can retrieve 8/bitspersample
 * pixel values simply by indexing into the table with one
 * number.
 */
static int
makecmap(TIFFRGBAImage * img)
{
	int bitspersample = img->bitspersample;
	int nsamples = 8 / bitspersample;
	uint16 * r = img->redcmap;
	uint16 * g = img->greencmap;
	uint16 * b = img->bluecmap;
	u_char * p;
	unsigned i;

	img->PALmap = (u_char**) _TIFFmalloc(256*sizeof(u_char *) + (256*nsamples*sizeof(u_char)));
	if (img->PALmap == NULL)
	{
		TIFFError(TIFFFileName(img->tif), "No space for Palette mapping table");
		return 0;
	}
	p = (u_char *) (img->PALmap + 256);
	for (i = 0; i < 256; i++)
	{
		img->PALmap[i] = p;
#define	CMAP(c)	*p++ = MapRGB(r[c]&0xFF, g[c]&0xFF, b[c]&0xFF);
		switch (bitspersample)
		{
		case 1:
			CMAP(i>>7);
			CMAP((i>>6)&0x01);
			CMAP((i>>5)&0x01);
			CMAP((i>>4)&0x01);
			CMAP((i>>3)&0x01);
			CMAP((i>>2)&0x01);
			CMAP((i>>1)&0x01);
			CMAP(i&0x01);
			break;
		case 2:
			CMAP(i>>6);
			CMAP((i>>4)&0x03);
			CMAP((i>>2)&0x03);
			CMAP(i&0x03);
			break;
		case 4:
			CMAP(i>>4);
			CMAP(i&0x0F);
			break;
		case 8:
			CMAP(i);
			break;
		}
#undef CMAP
	}
	return 1;
}

/* 
 * Construct any mapping table used
 * by the associated put routine.
 */
static int
buildMap(TIFFRGBAImage* img)
{
	switch (img->photometric)
	{
	case PHOTOMETRIC_RGB:
	case PHOTOMETRIC_YCBCR:
	case PHOTOMETRIC_SEPARATED:
		if (img->bitspersample == 8)
			break;
		/* fall thru... */
	case PHOTOMETRIC_MINISBLACK:
	case PHOTOMETRIC_MINISWHITE:
		if (!setupMap(img))
			return 0;
		break;
	case PHOTOMETRIC_PALETTE:
		/*
		* Convert 16-bit colormap to 8-bit (unless it looks
		* like an old-style 8-bit colormap).
		*/
		if (checkcmap(img) == 16)
			cvtcmap(img);
		else
			TIFFWarning(TIFFFileName(img->tif), "Assuming 8-bit colormap");
		/*
		* Use mapping table and colormap to construct
		* unpacking tables for samples < 8 bits.
		*/
		if (img->bitspersample <= 8 && !makecmap(img))
			return 0;
		break;
	}
	return 1;
}

/*
 * Select the appropriate conversion routine for packed data.
 */
static int
pickTileContigCase(TIFFRGBAImage* img)
{
	tileContigRoutine put = NULL;
	
	if (buildMap(img))
	{
		switch (img->photometric)
		{
		case PHOTOMETRIC_RGB:
			switch (img->bitspersample)
			{
			case 8:
				if (!img->Map)
				{
					if (img->alpha == EXTRASAMPLE_ASSOCALPHA)
						put = putRGBAAcontig8bittile;
					else if (img->alpha == EXTRASAMPLE_UNASSALPHA)
						put = putRGBUAcontig8bittile;
					else
						put = putRGBcontig8bittile;
				}
				else
					put = putRGBcontig8bitMaptile;
				break;
			case 16:
				put = putRGBcontig16bittile;
				if (!img->Map)
				{
					if (img->alpha == EXTRASAMPLE_ASSOCALPHA)
						put = putRGBAAcontig16bittile;
					else if (img->alpha == EXTRASAMPLE_UNASSALPHA)
						put = putRGBUAcontig16bittile;
				}
				break;
			}
			break;
		case PHOTOMETRIC_SEPARATED:
			if (img->bitspersample == 8)
			{
				if (!img->Map)
					put = putRGBcontig8bitCMYKtile;
				else
					put = putRGBcontig8bitCMYKMaptile;
			}
			break;
		case PHOTOMETRIC_PALETTE:
			switch (img->bitspersample)
			{
			case 8: put = put8bitcmaptile; break;
			case 4: put = put4bitcmaptile; break;
			case 2: put = put2bitcmaptile; break;
			case 1: put = put1bitcmaptile; break;
			}
			break;
		case PHOTOMETRIC_MINISWHITE:
		case PHOTOMETRIC_MINISBLACK:
			switch (img->bitspersample)
			{
			case 8: put = putgreytile; break;
			case 4: put = put4bitbwtile; break;
			case 2: put = put2bitbwtile; break;
			case 1: put = put1bitbwtile; break;
			}
			break;
#if defined(YCBCR_SUPPORTED)
		case PHOTOMETRIC_YCBCR:
			if (img->bitspersample == 8)
				put = initYCbCrConversion(img);
			break;
#endif
		}
	}
	return ((img->put.contig = put) != NULL);
}

/*
 * Select the appropriate conversion routine for unpacked data.
 *
 * NB: we assume that unpacked single channel data is directed
 *	 to the "packed" routines.
 */
static int
pickTileSeparateCase(TIFFRGBAImage* img)
{
	tileSeparateRoutine put = NULL;

	if (buildMap(img))
	{
		switch (img->photometric)
		{
		case PHOTOMETRIC_RGB:
			switch (img->bitspersample)
			{
			case 8:
				if (!img->Map)
				{
					if (img->alpha == EXTRASAMPLE_ASSOCALPHA)
						put = putRGBAAseparate8bittile;
					else if (img->alpha == EXTRASAMPLE_UNASSALPHA)
						put = putRGBUAseparate8bittile;
					else
						put = putRGBseparate8bittile;
				}
				else
					put = putRGBseparate8bitMaptile;
				break;
			case 16:
				put = putRGBseparate16bittile;
				if (!img->Map)
				{
					if (img->alpha == EXTRASAMPLE_ASSOCALPHA)
						put = putRGBAAseparate16bittile;
					else if (img->alpha == EXTRASAMPLE_UNASSALPHA)
						put = putRGBUAseparate16bittile;
				}
				break;
			}
			break;
		}
	}
	return ((img->put.separate = put) != NULL);
}
